/**
 * Author: Valerio Schiavoni <valerio.schiavoni@gmail.com>
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.rmi.Remote;
import java.util.Map;
import java.util.logging.Logger;

import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.bf.AbstractBindingFactoryPlugin;
import org.objectweb.fractal.bf.BindingFactoryException;
import org.objectweb.fractal.bf.CannotCreateSkeletonException;
import org.objectweb.fractal.bf.proxies.StubComponentFactory;
import org.objectweb.fractal.util.Fractal;

/**
 * 
 */
public class RmiConnector extends AbstractBindingFactoryPlugin<RmiExportId> {

	Logger log = Logger.getLogger(RmiConnector.class.getCanonicalName());

	public static Remote skeletonRef;

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#createExportId(java.lang.String,
	 *      org.objectweb.fractal.api.Component, java.util.Map)
	 */
	public RmiExportId createExportId(String interfaceName, Component owner,
			Map<String, String> hints) throws BindingFactoryException {
		String ownerName = null;
		try {
			ownerName = Fractal.getNameController(owner).getFcName();
		} catch (NoSuchInterfaceException e) {
			throw new BindingFactoryException(
					"Cannot get the name of the component owning the interface being exported with the SOAP-CXF connector",
					e);
		}
		return new RmiExportId(interfaceName, ownerName);
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#createSkel(java.lang.String,
	 *      java.lang.Object, org.objectweb.fractal.api.Component,
	 *      java.util.Map)
	 */
	public Component createSkel(String interfaceName, Object serverItf,
			Component owner, Map<String, String> exportHints)
			throws CannotCreateSkeletonException {
		log.info("RmiConnector - Creating skeleton..");

		String name = exportHints.get(RmiConnectorConstants.SERVICE_NAME);
		// a fractal component whose content is a rmi skeleton.
		Component compRmiSkel = null;

		try {

			Factory adlFactory = FactoryFactory
					.getFactory(FactoryFactory.FRACTAL_BACKEND);

			compRmiSkel = (Component) adlFactory.newComponent(
					"org.objectweb.fractal.bf.connectors.rmi.RmiServerBridge",
					null);

			// start the skeleton
			Fractal.getLifeCycleController(compRmiSkel).startFc();

			RmiConnectorAttributes rmiServiceAttributes = (RmiConnectorAttributes) compRmiSkel
					.getFcInterface(RmiConnectorAttributes.RMI_CONNECTOR_ATTRIBUTES);

			/*
			 * this should not be necessary if the ADL factory could take into
			 * account instance of parameter objects
			 */
			rmiServiceAttributes.setServiceObject(serverItf);
			rmiServiceAttributes.setServiceName(name);
			rmiServiceAttributes.setHost(exportHints
					.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS));
			rmiServiceAttributes.setPort(exportHints
					.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_PORT));

		} catch (Exception e) {

			e.printStackTrace();
			throw new CannotCreateSkeletonException(
					"Some error occurred while exporting using the JavaRMI plugin ",
					e);
		}

		return compRmiSkel;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#finalizeSkeleton(org.objectweb.fractal.api.Component,
	 *      java.util.Map)
	 */
	public void finalizeSkeleton(Component skel,
			Map<String, String> finalizationHints)
			throws BindingFactoryException {

		log.warning("FinalizationHints map currently ignored");

		try {
			RmiConnectorAttributes rca = (RmiConnectorAttributes) skel
					.getFcInterface(RmiConnectorAttributes.RMI_CONNECTOR_ATTRIBUTES);

			rca.startRmi();

		} catch (NoSuchInterfaceException e) {
			throw new BindingFactoryException(e);
		}

	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getStubComponentFactory()
	 */
	public StubComponentFactory getStubComponentFactory() {
		return new RmiConnectorStubComponentFactoryImpl();
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getStubGenerationParameters(java.util.Map)
	 */
	public String[] getStubGenerationParameters(Map<String, String> bindHints)
			throws BindingFactoryException {
		if (bindHints == null) {
			throw new BindingFactoryException("Binding hints were null");
		}
		String name = bindHints.get(RmiConnectorConstants.SERVICE_NAME);
		if (name == null) {
			throw new BindingFactoryException(
					"Missing parameter for JavaRmi plugin (expected:"
							+ RmiConnectorConstants.SERVICE_NAME + ")");
		}
		String registryHost = bindHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS);
		if (registryHost == null) {
			log
					.warning("You didn't specify "
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS
							+ ", will use '"
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT
							+ "' ");
			registryHost = RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT;
		}

		String registryPort = bindHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_PORT);
		if (registryPort == null) {
			log
					.warning("You didn't specify the port for the RMI registry, the default port [1099] will be used");
			registryPort = "1099";
		}

		String[] params = new String[] { name, registryHost, registryPort };
		return params;

	}
}
