/**
 * Author: Valerio Schiavoni <valerio.schiavoni@gmail.com>
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.lang.reflect.Constructor;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Arrays;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.bf.AbstractSkeleton;
import org.objectweb.fractal.bf.connectors.rmi.asm.RmiEnablingInterfaceTransformer;
import org.objectweb.fractal.bf.connectors.rmi.asm.UnicastRemoteObjectGenerator;

/**
 * 
 */
public class RmiServerBridge extends AbstractSkeleton implements
		RmiConnectorAttributes {

	Logger log = Logger.getLogger(RmiServerBridge.class.getCanonicalName());

	private Object service;
	private String name;
	private String host;
	private String port;

	public RmiServerBridge() {

	}

	public void startRmi() {
		Remote rmiSkeleton = null;
		Component interfaceOwner = ((Interface) service).getFcItfOwner();
		Object content = null;
		try {
			content = interfaceOwner.getFcInterface("/content");
		} catch (NoSuchInterfaceException e2) {
		}

		try {
			if (Remote.class.isAssignableFrom(service.getClass())) {
				log.fine("Fractal interface " + service
						+ " is a subclass of java.rmi.Remote");

				rmiSkeleton = UnicastRemoteObject.exportObject(
						(Remote) interfaceOwner.getFcInterface("/content"), 0);
			} else {
				log
						.fine("Activating runtime skeleton generation for interface: "
								+ service);
				RmiEnablingInterfaceTransformer transformer = new RmiEnablingInterfaceTransformer();
				Class<?> toTransform = null;
				Class<?>[] interfaces = content.getClass().getInterfaces();
				log.finest("Interfaces implemented by /content: "
						+ Arrays.deepToString(interfaces));
				toTransform = interfaces[0]; // XXX
				log.finest("Runtime class of interface to transform: "
						+ toTransform);
				final Class<?> remotedJavaInterface = transformer
						.transform(toTransform);
				log
						.finer("Corresponding java.rmi.Remote enabled interface for fractal interface created with name: "
								+ remotedJavaInterface.getCanonicalName());
				UnicastRemoteObjectGenerator generator = new UnicastRemoteObjectGenerator();

				Class<?> remotableClassForSkeleton = generator.generate(
						toTransform, remotedJavaInterface);
				log.finest("Subclass of UnicastRemoteObject generated: "
						+ remotableClassForSkeleton);
				Constructor<?> constructor;

				try {
					constructor = remotableClassForSkeleton
							.getConstructor(new Class[] { toTransform });
					log.finest("Constructor found: " + constructor);
					final Object newInstance = constructor
							.newInstance(new Object[] { content });
					log.finest("Skeleton class instantiated: " + newInstance);
					UnicastRemoteObject uro = (UnicastRemoteObject) newInstance;
					UnicastRemoteObject.unexportObject(uro, true);
					log.finest("Unexported...going to re-export it.");
					/* export it in the rmi registry */
					rmiSkeleton = UnicastRemoteObject
							.exportObject(
									(UnicastRemoteObject) newInstance,
									Integer
											.valueOf(RmiConnectorConstants.JAVA_RMI_REGISTRY_DEFAULT_PORT));
					log.finest("Rmi skeleton exported " + rmiSkeleton);

				} catch (Exception e) {
					e.printStackTrace();
					log
							.warning("Could not generate RMI-fied version of the business interface to export via RMI, due to: "
									+ e.getLocalizedMessage());
				}

			}
			Registry registry = locateRegistry();
			log.fine("Registry located: " + registry);

			registry.rebind(name, rmiSkeleton);
			log.fine("Skeleton bound into registry with name: " + name);

		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (NoSuchInterfaceException e) {

		}

	}

	/**
	 * @return a local reference to the RMI registry.
	 * @throws RemoteException
	 * @throw {@link RmiRegistryCreationException} if problems occur while
	 *        creating a local rmi registry.
	 */
	private Registry locateRegistry() {
		Integer portIntValue = Integer.valueOf(port);
		if (portIntValue == null) {
			log.warning("No port set for RMI registry, will use :"
					+ RmiConnectorConstants.JAVA_RMI_REGISTRY_DEFAULT_PORT);
			portIntValue = Integer
					.valueOf(RmiConnectorConstants.JAVA_RMI_REGISTRY_DEFAULT_PORT);
		}
		Registry registry = null;
		try {

			registry = LocateRegistry.getRegistry(host, portIntValue);
			log.fine("RMI registry recovered :" + registry.toString());
		} catch (Exception e) {
			// create it locally
			log
					.warning("Could not recover a local RMI registry, going to create a new one");
			try {
				registry = LocateRegistry.createRegistry(portIntValue);
			} catch (RemoteException e1) {
				log
						.warning("Could not create a local instance of RMI registry due to: "
								+ e1.getMessage());
				throw new RmiRegistryCreationException(
						"Cannot create a local reference for the rmi registry",
						e1);
			}
		}

		return registry;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setServiceName(java.lang.String)
	 */
	public void setServiceName(String name) {
		this.name = name;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setServiceObject(java.lang.Object)
	 */
	public void setServiceObject(Object service) {
		this.service = service;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setHost(java.lang.String)
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * @see org.objectweb.fractal.bf.connectors.rmi.RmiConnectorAttributes#setPort(java.lang.String)
	 */
	public void setPort(String port) {
		this.port = port;
	}

}
