/**
 * Author: Valerio Schiavoni <valerio.schiavoni@gmail.com>
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.lang.reflect.Method;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.logging.Logger;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * A simple class transformer that given a Java interface creates a
 * corresponding Class object that can be exported via Java RMI. The resulting
 * class extends {@link Remote} and for every method declared by the interface,
 * a {@link RemoteException} is added to the signature.
 * 
 * This code is losely based on the corresponding solution adopted in the
 * Tuscany source:
 * https://svn.apache.org/repos/asf/incubator/tuscany/java/sca/modules/binding-rmi/src/main/java/org/apache/tuscany/sca/binding/rmi/RMIService.java
 * 
 * The solution adopted there is to use a mix of ASM and CGLIB bytecode
 * instrumentation tools. Instead, this class uses only ASM.
 * 
 */
public class RmiEnablingInterfaceTransformer {
	/**
	 * The suffix appended to the internal binary name of the generated class.
	 */
	private static final String GENERATED_CLASS_SUFFIX = "Remoted";
	/* The logger to use for this class */
	Logger log = Logger.getLogger(RmiEnablingInterfaceTransformer.class
			.getCanonicalName());

	/**
	 * Transform a java interface to a corresponding java interface extending
	 * {@link java.rmi.Remote}.
	 * 
	 * @param targetJavaInterface
	 *            the java interface to be transformed
	 * @return a corresponding java interface ready to be exported with RMI, or
	 *         targetJavaInterface if it's already a subclass of java.rmi.Remote
	 */
	public Class<?> transform(Class<?> targetJavaInterface) {
		final String javaItfCanonicalName = targetJavaInterface
				.getCanonicalName();

		if (targetJavaInterface.isInterface()
				&& !Remote.class.isAssignableFrom(targetJavaInterface)) {
			log.finer("Target java interface " + javaItfCanonicalName
					+ " is not a subclass of java.rmi.Remote");

			// ClassLoader originalJavaInterfaceCL = targetJavaInterface
			// .getClassLoader();
			// log.finest("Original java interface class loader: "
			// + originalJavaInterfaceCL.toString());

			RMIServiceClassLoader classloader = RMIServiceClassLoaderFactory
					.getRMIClassLoader(getClass().getClassLoader());// new
			// RMIServiceClassLoader(
			// originalJavaInterfaceCL);
			log.finer("JavaRMI class loader: "
					+ RMIServiceClassLoader.class.getName());
			log.finest("Ref: " + classloader);

			final byte[] byteCode = generateRemoteInterface(targetJavaInterface);
			log.finest("Bytecode for RMI-enabled interface generated");
			log.finest("Original java interface name: " + javaItfCanonicalName);
			final String generatedInterfaceInternalName = javaItfCanonicalName
					+ GENERATED_CLASS_SUFFIX;
			log.finest("Generated java interface name: "
					+ generatedInterfaceInternalName);
			Class<?> remotedTargetJavaInterface = classloader.defineClass(
					byteCode, generatedInterfaceInternalName);
			log.finest("Class " + remotedTargetJavaInterface.getCanonicalName()
					+ " for RMI-enabled interface defined in the "
					+ RMIServiceClassLoader.class.getName());
			return remotedTargetJavaInterface;
		} else if (!targetJavaInterface.isInterface()) {
			log.warning("The class: " + targetJavaInterface
					+ " is not an interface");
		}
		log
				.finer("Target java interface: "
						+ javaItfCanonicalName
						+ " is a java.rmi.Remote subclass, no need to generate subclasses");
		return targetJavaInterface;
	}

	/**
	 * if the interface of the component whose serviceBindings must be exposed
	 * as RMI Service, does not implement java.rmi.Remote, then generate such an
	 * interface. This method will stop with just generating the bytecode.
	 * Defining the class from the byte code must be the responsibility of the
	 * caller of this method, since it requires a ClassLoader to be created to
	 * define and load this interface.
	 */
	protected byte[] generateRemoteInterface(Class<?> serviceInterface) {
		String interfazeName = serviceInterface.getCanonicalName();
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);

		final String internalNameGeneratedClass = interfazeName.replace('.',
				'/')
				+ GENERATED_CLASS_SUFFIX;
		log.finest("Internal name for generated class: "
				+ internalNameGeneratedClass);
		cw.visit(Opcodes.V1_5, Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT
				+ Opcodes.ACC_INTERFACE, internalNameGeneratedClass, null,
				"java/lang/Object", new String[] { "java/rmi/Remote",
						Type.getInternalName(serviceInterface) });

		StringBuffer argsAndReturn = null;
		Method[] methods = serviceInterface.getMethods();
		for (Method method : methods) {
			argsAndReturn = new StringBuffer("(");
			Class<?>[] paramTypes = method.getParameterTypes();
			Class<?> returnType = method.getReturnType();

			for (Class<?> paramType : paramTypes) {
				argsAndReturn.append(Type.getType(paramType));
			}
			argsAndReturn.append(")");
			argsAndReturn.append(Type.getType(returnType));

			cw.visitMethod(Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT, method
					.getName(), argsAndReturn.toString(), null,
					new String[] { "java/rmi/RemoteException" });
		}
		cw.visitEnd();
		return cw.toByteArray();
	}

}
